@@ -1,4 +1,4 @@
-# Copyright 2004-2016 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -24,6 +24,8 @@
 # their behavior, while functions imported in are probably best left
 # alone as part of the api.
 
+from __future__ import print_function
+
 # Remember the real file.
 _file = file
 
@@ -34,6 +36,7 @@ import renpy.audio
 
 from renpy.pyanalysis import const, pure, not_const
 
+
 def renpy_pure(fn):
     """
     Marks renpy.`fn` as a pure function.
@@ -54,6 +57,7 @@ from renpy.text.text import language_tailor
 from renpy.display.behavior import Keymap
 from renpy.display.behavior import run, run as run_action, run_unhovered, run_periodic
 from renpy.display.behavior import map_event, queue_event, clear_keymap_cache
+from renpy.display.behavior import is_selected, is_sensitive
 
 from renpy.display.minigame import Minigame
 from renpy.display.screen import define_screen, show_screen, hide_screen, use_screen, current_screen
@@ -64,7 +68,7 @@ from renpy.display.focus import focus_coordinates
 from renpy.display.predict import screen as predict_screen
 
 from renpy.display.image import image_exists, image_exists as has_image
-from renpy.display.image import get_available_image_tags, get_available_image_attributes
+from renpy.display.image import get_available_image_tags, get_available_image_attributes, check_image_attributes, get_ordered_image_attributes
 
 from renpy.display.im import load_surface, load_image
 
@@ -115,6 +119,7 @@ import time
 import sys
 import threading
 
+
 def public_api():
     """
     :undocumented:
@@ -152,7 +157,7 @@ def public_api():
     map_event, queue_event, clear_keymap_cache
     const, pure, not_const
     image_exists, has_image
-    get_available_image_tags, get_available_image_attributes
+    get_available_image_tags, get_available_image_attributes, check_image_attributes, get_ordered_image_attributes
     load_image, load_surface
     profile_memory, diff_memory, profile_rollback
     TEXT_TAG
@@ -162,6 +167,7 @@ def public_api():
     not_infinite_loop
     register_sl_statement, register_sl_displayable
     eval_who
+    is_selected, is_sensitive
 
 del public_api
 
@@ -301,6 +307,7 @@ def retain_after_load():
 
 scene_lists = renpy.display.core.scene_lists
 
+
 def count_displayables_in_layer(layer):
     """
     Returns how many displayables are in the supplied layer.
@@ -459,6 +466,7 @@ def showing(name, layer='master'):
 
     return renpy.game.context().images.showing(layer, name)
 
+
 def get_showing_tags(layer='master'):
     """
     :doc: image_func
@@ -468,6 +476,7 @@ def get_showing_tags(layer='master'):
 
     return renpy.game.context().images.get_showing_tags(layer)
 
+
 def predict_show(name, layer=None, what=None, tag=None, at_list=[ ]):
     """
     :undocumented:
@@ -521,6 +530,8 @@ def predict_show(name, layer=None, what=None, tag=None, at_list=[ ]):
         else:
             img = i(img)
 
+        img._unique()
+
     renpy.game.context().images.predict_show(layer, name, True)
     renpy.display.predict.displayable(img)
 
@@ -556,7 +567,7 @@ def show(name, at_list=[ ], layer=None, what=None, zorder=None, tag=None, behind
         zorder is preserved if it exists, and is otherwise set to 0.
 
     `tag`
-        A string, used to specify the the image tag of the shown image. The
+        A string, used to specify the image tag of the shown image. The
         equivalent of the ``as`` property.
 
     `behind`
@@ -633,6 +644,9 @@ def show(name, at_list=[ ], layer=None, what=None, zorder=None, tag=None, behind
         else:
             img = i(img)
 
+        # Mark the newly created images unique.
+        img._unique()
+
     # Update the list of images we have ever seen.
     renpy.game.persistent._seen_images[name] = True  # @UndefinedVariable
 
@@ -708,7 +722,7 @@ def scene(layer='master'):
         renpy.config.missing_scene(layer)
 
 
-def input(prompt, default='', allow=None, exclude='{}', length=None, with_none=None, pixel_width=None): #@ReservedAssignment
+def input(prompt, default='', allow=None, exclude='{}', length=None, with_none=None, pixel_width=None):  # @ReservedAssignment
     """
     :doc: input
 
@@ -748,7 +762,7 @@ def input(prompt, default='', allow=None, exclude='{}', length=None, with_none=N
     if roll_forward is not None:
         default = roll_forward
 
-    fixed = in_fixed_rollback();
+    fixed = in_fixed_rollback()
 
     if has_screen("input"):
         widget_properties = { }
@@ -807,7 +821,6 @@ def menu(items, set_expr):
         def substitute(s):
             return s
 
-
     # Filter the list of items to only include ones for which the
     # condition is true.
     items = [ (substitute(label), value)
@@ -816,12 +829,12 @@ def menu(items, set_expr):
 
     # Filter the list of items on the set_expr:
     if set_expr:
-        set = renpy.python.py_eval(set_expr) #@ReservedAssignment
+        set = renpy.python.py_eval(set_expr)  # @ReservedAssignment
         items = [ (label, value)
                   for label, value in items
                   if label not in set ]
     else:
-        set = None #@ReservedAssignment
+        set = None  # @ReservedAssignment
 
     # Check to see if there's at least one choice in set of items:
     choices = [ value for label, value in items if value is not None ]
@@ -864,7 +877,6 @@ def choice_for_skipping():
         renpy.loadsave.force_autosave(True)
 
 
-
 def predict_menu():
     """
     :undocumented:
@@ -888,6 +900,12 @@ def predict_menu():
         )
 
 
+class MenuEntry(tuple):
+    """
+    The object passed into the choice screen.
+    """
+
+
 def display_menu(items,
                  window_style='menu_window',
                  interact=True,
@@ -900,7 +918,7 @@ def display_menu(items,
                  scope={ },
                  widget_properties=None,
                  screen="choice",
-                 type="menu", #@ReservedAssignment
+                 type="menu",  # @ReservedAssignment
                  predict_only=False,
                  **kwargs):
     """
@@ -968,16 +986,21 @@ def display_menu(items,
 
             if value is not None:
                 action = renpy.ui.ChoiceReturn(label, value, location)
+                chosen = action.get_chosen()
             else:
                 action = None
+                chosen = False
 
             if renpy.config.choice_screen_chosen:
-                if action is not None:
-                    item_actions.append((label, action, action.get_chosen()))
-                else:
-                    item_actions.append((label, action, False))
+                me = MenuEntry((label, action, chosen))
             else:
-                item_actions.append((label, action))
+                me = MenuEntry((label, action))
+
+            me.caption = label
+            me.action = action
+            me.chosen = chosen
+
+            item_actions.append(me)
 
             show_screen(screen, items=item_actions, _widget_properties=props, _transient=True, _layer=renpy.config.choice_layer, **scope)
 
@@ -1032,6 +1055,7 @@ def display_menu(items,
 
 
 class TagQuotingDict(object):
+
     def __getitem__(self, key):
 
         store = renpy.store.__dict__
@@ -1059,7 +1083,7 @@ def predict_say(who, what):
     """
 
     if who is None:
-        who = renpy.store.narrator # E1101 @UndefinedVariable
+        who = renpy.store.narrator  # E1101 @UndefinedVariable
 
     if isinstance(who, (str, unicode)):
         return renpy.store.predict_say(who, what)
@@ -1082,6 +1106,7 @@ def scry_say(who, scry):
     except:
         scry.interacts = True
 
+
 def say(who, what, interact=True):
     """
     :doc: se_say
@@ -1114,7 +1139,7 @@ def say(who, what, interact=True):
         what = what % tag_quoting_dict
 
     if who is None:
-        who = renpy.store.narrator # E1101 @UndefinedVariable
+        who = renpy.store.narrator  # E1101 @UndefinedVariable
 
     if isinstance(who, (str, unicode)):
         renpy.store.say(who, what, interact=interact)
@@ -1210,10 +1235,9 @@ def pause(delay=None, music=None, with_none=None, hard=False, checkpoint=None):
     `checkpoint`
         If true, a checkpoint will be set, and players will be able to roll
         back to this statement. If false, no checkpoint will be set. If None,
-        a checkpoint will only be set if display is set.
+        a checkpoint will only be set if delay is set.
     """
 
-
     if checkpoint is None:
         if delay is not None:
             checkpoint = False
@@ -1359,7 +1383,7 @@ def with_statement(trans, always=False, paired=None, clear=True):
 globals()["with"] = with_statement
 
 
-def rollback(force=False, checkpoints=1, defer=False, greedy=True, label=None):
+def rollback(force=False, checkpoints=1, defer=False, greedy=True, label=None, abnormal=True):
     """
     :doc: rollback
 
@@ -1384,6 +1408,11 @@ def rollback(force=False, checkpoints=1, defer=False, greedy=True, label=None):
 
     `label`
         If not None, a label that is called when rollback completes.
+
+    `abnormal`
+        If true, the default, the code executed after the transition is run in
+        an abnormal mode that skips transitions that would have otherwise
+        occured.
     """
 
     if defer and len(renpy.game.contexts) > 1:
@@ -1403,7 +1432,7 @@ def rollback(force=False, checkpoints=1, defer=False, greedy=True, label=None):
 
     renpy.config.skipping = None
     renpy.game.log.complete()
-    renpy.game.log.rollback(checkpoints, greedy=greedy, label=label, force=force)
+    renpy.game.log.rollback(checkpoints, greedy=greedy, label=label, force=(force is True), abnormal=abnormal)
 
 
 def toggle_fullscreen():
@@ -1421,6 +1450,7 @@ def toggle_music():
     Does nothing.
     """
 
+
 @renpy_pure
 def has_label(name):
     """
@@ -1435,6 +1465,7 @@ def has_label(name):
 
     return renpy.game.script.has_label(name)
 
+
 @renpy_pure
 def get_all_labels():
     """
@@ -1504,13 +1535,20 @@ def reload_script():
 
     s = get_screen("menu")
 
-    if s is not None:
-        session["_reload_screen"] = s.screen_name[0]
+    if not renpy.store.main_menu:
+        if s is not None:
+            session["_reload_screen"] = s.screen_name[0]
 
-    renpy.game.call_in_new_context("_save_reload_game")
+        renpy.game.call_in_new_context("_save_reload_game")
+
+    else:
+        if s is not None:
+            session["_main_menu_screen"] = s.screen_name[0]
+
+        utter_restart()
 
 
-def quit(relaunch=False, status=0): #@ReservedAssignment
+def quit(relaunch=False, status=0):  # @ReservedAssignment
     """
     :doc: other
 
@@ -1560,6 +1598,7 @@ def call(label, *args, **kwargs):
 
     raise renpy.game.CallException(label, args, kwargs)
 
+
 def return_statement():
     """
     :doc: se_call
@@ -1579,8 +1618,9 @@ def screenshot(filename):
 
     renpy.game.interface.save_screenshot(filename)
 
+
 @renpy_pure
-def version(tuple=False): #@ReservedAssignment
+def version(tuple=False):  # @ReservedAssignment
     """
     :doc: renpy_version
 
@@ -1600,7 +1640,8 @@ version_string = renpy.version
 version_only = renpy.version_only
 version_name = renpy.version_name
 version_tuple = renpy.version_tuple
-license = "" # @ReservedAssignment
+license = ""  # @ReservedAssignment
+
 
 def transition(trans, layer=None, always=False, force=False):
     """
@@ -1660,6 +1701,7 @@ def get_game_runtime():
 
     return renpy.game.contexts[0].runtime
 
+
 @renpy_pure
 def loadable(filename):
     """
@@ -1775,6 +1817,7 @@ def get_filename_line():
 # A file that log logs to.
 logfile = None
 
+
 def log(msg):
     """
     :doc: debug
@@ -1803,7 +1846,7 @@ def log(msg):
 
         import textwrap
 
-        print >>logfile, textwrap.fill(msg, renpy.config.log_width).encode("utf-8")
+        print(textwrap.fill(msg, renpy.config.log_width).encode("utf-8"), file=logfile)
         logfile.flush()
 
     except:
@@ -1850,7 +1893,7 @@ def last_interact_type():
     return getattr(renpy.game.context().info, "_last_interact_type", None)
 
 
-def dynamic(*vars): #@ReservedAssignment
+def dynamic(*vars):  # @ReservedAssignment
     """
     :doc: other
 
@@ -1866,7 +1909,7 @@ def dynamic(*vars): #@ReservedAssignment
     renpy.game.context().make_dynamic(vars)
 
 
-def context_dynamic(*vars): #@ReservedAssignment
+def context_dynamic(*vars):  # @ReservedAssignment
     """
     :doc: other
 
@@ -1922,7 +1965,7 @@ def seen_image(name):
     return name in renpy.game.persistent._seen_images  # @UndefinedVariable
 
 
-def file(fn): #@ReservedAssignment
+def file(fn):  # @ReservedAssignment
     """
     :doc: file
 
@@ -2007,6 +2050,7 @@ def free_memory():
     force_full_redraw()
     renpy.display.interface.kill_textures_and_surfaces()
 
+
 @renpy_pure
 def easy_displayable(d, none=False):
     """
@@ -2056,12 +2100,15 @@ def _error(msg):
 
 _error_handlers = [ _error ]
 
+
 def push_error_handler(eh):
     _error_handlers.append(eh)
 
+
 def pop_error_handler():
     _error_handlers.pop()
 
+
 def error(msg):
     _error_handlers[-1](msg)
 
@@ -2114,6 +2161,7 @@ def scry():
     node = renpy.game.script.lookup(name)
     return node.scry()
 
+
 @renpy_pure
 def munged_filename():
     return renpy.parser.munge_filename(get_filename_line()[0])
@@ -2123,6 +2171,7 @@ def munged_filename():
 
 loaded_modules = set()
 
+
 def load_module(name, **kwargs):
     """
     :doc: other
@@ -2156,9 +2205,9 @@ def load_module(name, **kwargs):
     renpy.game.contexts.append(context)
 
     context.make_dynamic(kwargs)
-    renpy.store.__dict__.update(kwargs) #@UndefinedVariable
+    renpy.store.__dict__.update(kwargs)  # @UndefinedVariable
 
-    for prio, node in initcode: #@UnusedVariable
+    for prio, node in initcode:  # @UnusedVariable
         renpy.game.context().run(node)
 
     context.pop_all_dynamic()
@@ -2196,7 +2245,7 @@ def load_string(s, filename="<string>"):
         context.init_phase = True
         renpy.game.contexts.append(context)
 
-        for prio, node in initcode: #@UnusedVariable
+        for prio, node in initcode:  # @UnusedVariable
             renpy.game.context().run(node)
 
         context.pop_all_dynamic()
@@ -2266,6 +2315,7 @@ def shown_window():
 
 
 class placement(renpy.python.RevertableObject):
+
     def __init__(self, p):
         super(placement, self).__init__()
 
@@ -2277,6 +2327,7 @@ class placement(renpy.python.RevertableObject):
         self.yoffset = p[5]
         self.subpixel = p[6]
 
+
 def get_placement(d):
     """
     :doc: image_func
@@ -2336,9 +2387,11 @@ render = renpy.display.render.render
 IgnoreEvent = renpy.display.core.IgnoreEvent
 redraw = renpy.display.render.redraw
 
+
 class Displayable(renpy.display.core.Displayable, renpy.python.RevertableObject):
     pass
 
+
 class Container(renpy.display.core.Displayable, renpy.python.RevertableObject):
     _list_type = renpy.python.RevertableList
 
@@ -2366,7 +2419,6 @@ def cache_pin(*args):
     renpy.store._cache_pin_set = new_pins | renpy.store._cache_pin_set
 
 
-
 def cache_unpin(*args):
     """
     :undocumented: Cache pin is deprecated.
@@ -2422,7 +2474,6 @@ def stop_predict(*args):
 
 
 def start_predict_screen(_screen_name, *args, **kwargs):
-
     """
     :doc: screens
 
@@ -2473,7 +2524,7 @@ def call_screen(_screen_name, *args, **kwargs):
 
     try:
         rv = renpy.ui.interact(mouse="screen", type="screen", roll_forward=roll_forward)
-    except (renpy.game.JumpException, renpy.game.CallException), e:
+    except (renpy.game.JumpException, renpy.game.CallException) as e:
         rv = e
 
     renpy.exports.checkpoint(rv)
@@ -2491,6 +2542,7 @@ def call_screen(_screen_name, *args, **kwargs):
 
     return rv
 
+
 @renpy_pure
 def list_files(common=False):
     """
@@ -2506,7 +2558,7 @@ def list_files(common=False):
 
     rv = [ ]
 
-    for dir, fn in renpy.loader.listdirfiles(common): #@ReservedAssignment
+    for dir, fn in renpy.loader.listdirfiles(common):  # @ReservedAssignment
         if fn.startswith("saves/"):
             continue
 
@@ -2580,6 +2632,7 @@ def mode(mode):
 
     modes.insert(0, mode)
 
+
 def get_mode():
     """
     :doc: modes
@@ -2596,6 +2649,7 @@ def get_mode():
 
     return modes[0]
 
+
 def notify(message):
     """
     :doc: other
@@ -2649,7 +2703,7 @@ def vibrate(duration):
     """
 
     try:
-        import android #@UnresolvedImport
+        import android  # @UnresolvedImport
         android.vibrate(duration)
     except:
         pass
@@ -2733,6 +2787,8 @@ def set_physical_size(size):
     if get_renderer_info()["resizable"]:
         renpy.display.draw.quit()
         renpy.display.interface.set_mode(size)
+        renpy.display.interface.last_resize = size
+
 
 def reset_physical_size():
     """
@@ -2782,7 +2838,7 @@ def fsdecode(s):
     return s.decode(fsencoding)
 
 
-from renpy.editor import launch_editor #@UnusedImport
+from renpy.editor import launch_editor  # @UnusedImport
 
 
 def get_image_load_log(age=None):
@@ -2811,7 +2867,6 @@ def get_image_load_log(age=None):
     else:
         deadline = 0
 
-
     for i in renpy.display.im.cache.load_log:
         if i[0] < deadline:
             break
@@ -2919,14 +2974,22 @@ def count_seen_dialogue_blocks():
 
     Returns the number of dialogue blocks the user has seen in any play-through
     of the current game.
-
-    This number may be larger than the result of :func:`renpy.count_dialogue_blocks`
-    when the script has changed and older dialogue blocks are no longer accessible.
     """
 
     return renpy.game.seen_translates_count
 
 
+def count_newly_seen_dialogue_blocks():
+    """
+    :doc: other
+
+    Returns the number of dialogue blocks the user has seen for the first time
+    during this session.
+    """
+
+    return renpy.game.new_translates_count
+
+
 def substitute(s, scope=None, translate=True):
     """
     :doc: other
@@ -2982,6 +3045,7 @@ def get_return_stack():
 
     return renpy.game.context().get_return_stack()
 
+
 def set_return_stack(stack):
     """
     :doc: label
@@ -2995,6 +3059,7 @@ def set_return_stack(stack):
 
     renpy.game.context().set_return_stack(stack)
 
+
 def invoke_in_thread(fn, *args, **kwargs):
     """
     :doc: other
@@ -3020,6 +3085,7 @@ def invoke_in_thread(fn, *args, **kwargs):
     t.daemon = True
     t.start()
 
+
 def cancel_gesture():
     """
     :doc: gesture
@@ -3028,7 +3094,8 @@ def cancel_gesture():
     This should be called by displayables that have gesture-like behavior.
     """
 
-    renpy.display.gesture.recognizer.cancel() # @UndefinedVariable
+    renpy.display.gesture.recognizer.cancel()  # @UndefinedVariable
+
 
 def execute_default_statement(start=False):
     """
@@ -3040,6 +3107,7 @@ def execute_default_statement(start=False):
     for i in renpy.ast.default_statements:
         i.set_default(start)
 
+
 def write_log(s, *args):
     """
     :undocumented:
@@ -3049,6 +3117,7 @@ def write_log(s, *args):
 
     renpy.display.log.write(s, *args)
 
+
 def predicting():
     """
     :doc: screens
@@ -3058,6 +3127,7 @@ def predicting():
 
     return renpy.display.predict.predicting
 
+
 def get_line_log():
     """
     :undocumented:
@@ -3068,6 +3138,7 @@ def get_line_log():
 
     return renpy.game.context().line_log[:]
 
+
 def clear_line_log():
     """
     :undocumented:
@@ -3077,6 +3148,7 @@ def clear_line_log():
 
     renpy.game.context().line_log = [ ]
 
+
 def add_layer(layer, above=None, below=None, menu_clear=True):
     """
     :doc: other
@@ -3128,7 +3200,7 @@ def add_layer(layer, above=None, below=None, menu_clear=True):
     layers.insert(index, layer)
 
     if menu_clear:
-        renpy.config.menu_clear_layers.append(layer) # @UndefinedVariable
+        renpy.config.menu_clear_layers.append(layer)  # @UndefinedVariable
 
 
 def maximum_framerate(t):
@@ -3177,6 +3249,7 @@ def play(filename, channel=None, **kwargs):
 
     renpy.audio.music.play(filename, channel=channel, **kwargs)
 
+
 def get_editable_input_value():
     """
     :undocumented:
@@ -3187,6 +3260,7 @@ def get_editable_input_value():
 
     return renpy.display.behavior.current_input_value, renpy.display.behavior.input_value_active
 
+
 def set_editable_input_value(input_value, editable):
     """
     :undocumented:
@@ -3197,3 +3271,63 @@ def set_editable_input_value(input_value, editable):
 
     renpy.display.behavior.current_input_value = input_value
     renpy.display.behavior.input_value_active = editable
+
+
+def get_refresh_rate(precision=5):
+    """
+    :doc: other
+
+    Returns the refresh rate of the current screen, as a floating-point
+    number of frames per second.
+
+    `precision`
+        The raw data Ren'Py gets is number of frames per second, rounded down.
+        This means that a monitor that runs at 59.95 frames per second will
+        be reported at 59 fps. The precision argument reduces the precision
+        of this reading, such that the only valid readings are multiples of
+        the precision.
+
+        Since all monitor framerates tend to be multiples of 5 (25, 30, 60,
+        75, and 120), this likely will improve accuracy. Setting precision
+        to 1 disables this.
+    """
+
+    precision *= 1.0
+
+    info = renpy.display.get_info()
+    rv = info.refresh_rate
+    rv = round(rv / precision) * precision
+
+    return rv
+
+
+def get_identifier_checkpoints(identifier):
+    """
+    :doc: rollback
+
+    Given a rollback_identifier from a HistoryEntry object, returns the number
+    of checkpoints that need to be passed to :func:`renpy.rollback` to reach
+    that identifier. Returns None of the identifier is not in the rollback
+    history.
+    """
+
+    return renpy.game.log.get_identifier_checkpoints(identifier)
+
+
+def get_adjustment(bar_value):
+    """
+    :doc: other
+
+    Given `bar_value`, a  :class:`BarValue`, returns the :func:`ui.adjustment`
+    if uses. The adjustment has the following to attributes defined:
+
+    .. attribute:: value
+
+        The current value of the bar.
+
+    .. attribute:: range
+
+        The current range of the bar.
+    """
+
+    return bar_value.get_adjustment()
